import configparser
import filecmp
import os
import shutil
from urllib.parse import urljoin

import requests
from enhydris_api_client import EnhydrisApiClient

from .meteologgerstorage import MeteologgerStorage_wdat5

wdat5_parameters = [x.split()[1] for x in MeteologgerStorage_wdat5.wdat_record_format][
    5:
]
convertable_parameters = wdat5_parameters + ["fields"]


class ConfigFile:
    def __init__(self, filename):
        self.filename = filename

    def upgrade(self):
        self._read_config()
        self.api_client = EnhydrisApiClient(self.base_url)
        self._convert_username_and_password_to_api_token()
        self._convert_fields()
        self._backup_file()
        self._write_upgraded_file()

    def _read_config(self):
        self.config = configparser.ConfigParser(interpolation=None)
        with open(self.filename) as f:
            self.config.read_file(f)
        self.base_url = self.config.get("General", "base_url")

    def _convert_username_and_password_to_api_token(self):
        username = self.config.get("General", "username")
        password = self.config.get("General", "password")
        token = self.api_client.get_token(username, password)
        self.config.set("General", "auth_token", token)
        self.config.remove_option("General", "username")
        self.config.remove_option("General", "password")

    def _convert_fields(self):
        station_section_names = [n for n in self.config.sections() if n != "General"]
        for section_name in station_section_names:
            self._convert_section(section_name)

    def _convert_section(self, section_name):
        self.section_name = section_name
        self.station_id = self.config.get(section_name, "station_id")
        for parameter in convertable_parameters:
            self._convert_parameter(parameter)

    def _convert_parameter(self, parameter):
        value = self.config.get(self.section_name, parameter, fallback=None)
        if value is None:
            return
        timeseries_ids = [int(x.strip()) for x in value.split(",")]
        timeseries_group_ids = [
            self._get_timeseries_group(timeseries_id)
            for timeseries_id in timeseries_ids
        ]
        self.config.set(
            self.section_name,
            parameter,
            ",".join([str(x) for x in timeseries_group_ids]),
        )

    def _get_timeseries_group(self, timeseries_id):
        if timeseries_id == 0:
            return 0
        url = urljoin(
            self.api_client.base_url,
            f"api/stations/{self.station_id}/timeseries/{timeseries_id}/",
        )
        response = self.api_client.session.get(url)
        self._check_response(response)
        return response.json()["timeseries_group"]

    def _check_response(self, response):
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if response.text:
                raise requests.HTTPError(f"{str(e)}. Server response: {response.text}")
            else:
                raise

    @property
    def backup_filename(self):
        return self.filename + ".bak"

    def _backup_file(self):
        backup_file_is_identical = os.path.exists(self.backup_filename) and filecmp.cmp(
            self.filename, self.backup_filename, shallow=False
        )
        if backup_file_is_identical:
            return
        if os.path.exists(self.backup_filename):
            raise RuntimeError(
                f"Cannot backup configuration file; {self.backup_filename} exists"
            )
        shutil.copy(self.filename, self.backup_filename)

    def _write_upgraded_file(self):
        with open(self.filename, "w") as f:
            self.config.write(f)



import os
import re
import sys

sys.path.insert(0, os.path.abspath(".."))


def get_version():
    scriptdir = os.path.dirname(os.path.abspath(__file__))
    init_py_path = os.path.join(scriptdir, "..", "loggertodb", "__init__.py")
    with open(init_py_path) as f:
        return re.search(r'^__version__ = "(.*?)"$', f.read(), re.MULTILINE).group(1)


extensions = ["sphinx.ext.autodoc", "sphinx.ext.viewcode"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "loggertodb"
copyright = "2004-2020 various entities"
author = "Antonis Christofides"
version = get_version()
release = version
language = None
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
todo_include_todos = False
html_theme = "alabaster"
html_static_path = ["_static"]
htmlhelp_basename = "loggertodbdoc"
latex_elements = {}
latex_documents = [
    (
        master_doc,
        "loggertodb.tex",
        "loggertodb Documentation",
        "Antonis Christofides",
        "manual",
    )
]
man_pages = [(master_doc, "loggertodb", "loggertodb Documentation", [author], 1)]
texinfo_documents = [
    (
        master_doc,
        "loggertodb",
        "loggertodb Documentation",
        author,
        "loggertodb",
        "One line description of project.",
        "Miscellaneous",
    )
]



import datetime as dt
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from loggertodb.enhydris import Enhydris


@patch("loggertodb.enhydris.HTimeseries", new=lambda x: x)
class UploadTestCase(TestCase):
    @patch("loggertodb.enhydris.EnhydrisApiClient")
    def setUp(self, mock_EnhydrisApiClient):
        self.EnhydrisApiClient = mock_EnhydrisApiClient
        self._configure_EnhydrisApiClient(
            "get_ts_end_date.return_value", dt.datetime(2019, 3, 5, 7, 20)
        )
        self.MeteologgerStorage = MagicMock()
        self.enhydris = Enhydris(MagicMock())

    def _configure_EnhydrisApiClient(self, attribute, value):
        self.EnhydrisApiClient.configure_mock(**{f"return_value.{attribute}": value})

    def _configure_MeteologgerStorage(self, attribute, value):
        self.MeteologgerStorage.configure_mock(**{f"return_value.{attribute}": value})

    def test_calls_list_timeseries_as_needed(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1, 2, 3})
        self._configure_EnhydrisApiClient("list_timeseries.return_value", [])
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.list_timeseries.assert_has_calls(
            [call(42, 1), call(42, 2), call(42, 3)]
        )

    def test_determines_correct_timeseries_from_timeseries_group(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1})
        self._configure_EnhydrisApiClient(
            "list_timeseries.return_value",
            [{"id": 4242, "type": "Initial"}, {"id": 4243, "type": "Checked"}],
        )
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.get_ts_end_date.assert_called_once_with(
            42, 1, 4242
        )

    def test_creates_timeseries_if_not_exists(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1})
        self._configure_EnhydrisApiClient(
            "list_timeseries.return_value", [{"id": 4242, "type": "Checked"}]
        )
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.post_timeseries.assert_called_once_with(
            42, 1, data={"type": "Initial", "time_step": "", "timeseries_group": 1}
        )

    def test_does_not_create_timeseries_if_exists(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1})
        self._configure_EnhydrisApiClient(
            "list_timeseries.return_value", [{"id": 4242, "type": "Initial"}]
        )
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.post_timeseries.assert_not_called()

    def test_uses_id_of_created_timeseries(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1})
        self._configure_MeteologgerStorage("get_recent_data.return_value", "new data")
        self._configure_EnhydrisApiClient(
            "list_timeseries.return_value", [{"id": 4242, "type": "Checked"}]
        )
        self._configure_EnhydrisApiClient("post_timeseries.return_value", 9876)
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.post_tsdata.assert_has_calls(
            [call(42, 1, 9876, "new data")]
        )

    def test_calls_get_recent_data_as_needed(self):
        self._configure_MeteologgerStorage("timeseries_group_ids", {1, 2, 3})
        self._configure_MeteologgerStorage(
            "get_recent_data.side_effect", ["irrelevant", "irrelevant", "irrelevant"]
        )
        self.enhydris.upload(self.MeteologgerStorage())
        self.MeteologgerStorage.return_value.get_recent_data.assert_has_calls(
            [
                call(1, dt.datetime(2019, 3, 5, 7, 20)),
                call(2, dt.datetime(2019, 3, 5, 7, 20)),
                call(3, dt.datetime(2019, 3, 5, 7, 20)),
            ]
        )

    def test_calls_post_tsdata_as_needed(self):
        self._configure_MeteologgerStorage("station_id", 42)
        self._configure_MeteologgerStorage("timeseries_group_ids", {1, 2})
        self._configure_MeteologgerStorage(
            "get_recent_data.side_effect",
            [
                "new data for timeseries_group_id=1",
                "new data for timeseries_group_id=2",
            ],
        )
        self._configure_EnhydrisApiClient(
            "list_timeseries.side_effect",
            [[{"id": 4242, "type": "Initial"}], [{"id": 4243, "type": "Initial"}]],
        )
        self.enhydris.upload(self.MeteologgerStorage())
        self.EnhydrisApiClient.return_value.post_tsdata.assert_has_calls(
            [
                call(42, 1, 4242, "new data for timeseries_group_id=1"),
                call(42, 2, 4243, "new data for timeseries_group_id=2"),
            ]
        )



import datetime as dt
import os
from unittest import TestCase
from unittest.mock import call, patch

from loggertodb.exceptions import ConfigurationError
from loggertodb.meteologgerstorage import MeteologgerStorage_wdat5


class ConfigurationTestCase(TestCase):
    def test_accepts_allowed_optional_parameters(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "outsidetemp": 1256,
                "hioutsidetemp": 1257,
                "rain": 1652,
            }
        )

    def test_timeseries_group_ids(self):
        meteologger_storage = MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "wdat5",
                "outsidetemp": 1256,
                "hioutsidetemp": 1257,
                "rain": 1652,
            }
        )
        self.assertEqual(meteologger_storage.timeseries_group_ids, {1256, 1257, 1652})

    def test_valid_temperature_unit(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "temperature_unit": "C",
            }
        )

    def test_invalid_temperature_unit(self):
        msg = "temperature_unit must be one of C, F"
        with self.assertRaisesRegex(ConfigurationError, msg):
            MeteologgerStorage_wdat5(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "simple",
                    "temperature_unit": "A",
                }
            )

    def test_valid_rain_unit(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "rain_unit": "mm",
            }
        )

    def test_invalid_rain_unit(self):
        msg = "rain_unit must be one of mm, inch"
        with self.assertRaisesRegex(ConfigurationError, msg):
            MeteologgerStorage_wdat5(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "simple",
                    "rain_unit": "A",
                }
            )

    def test_valid_wind_speed_unit(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "wind_speed_unit": "m/s",
            }
        )

    def test_invalid_wind_speed_unit(self):
        msg = "wind_speed_unit must be one of m/s, mph"
        with self.assertRaisesRegex(ConfigurationError, msg):
            MeteologgerStorage_wdat5(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "simple",
                    "wind_speed_unit": "A",
                }
            )

    def test_valid_pressure_unit(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "pressure_unit": "hPa",
            }
        )

    def test_invalid_pressure_unit(self):
        msg = "pressure_unit must be one of hPa, inch Hg"
        with self.assertRaisesRegex(ConfigurationError, msg):
            MeteologgerStorage_wdat5(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "simple",
                    "pressure_unit": "A",
                }
            )

    def test_valid_matric_potential_unit(self):
        MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "matric_potential_unit": "centibar",
            }
        )

    def test_invalid_matric_potential_unit(self):
        msg = "matric_potential_unit must be one of centibar, cm"
        with self.assertRaisesRegex(ConfigurationError, msg):
            MeteologgerStorage_wdat5(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "simple",
                    "matric_potential_unit": "A",
                }
            )


class GetStorageTailTestCase(TestCase):
    @classmethod
    @patch(
        "loggertodb.meteologgerstorage.MeteologgerStorage._fix_dst",
        side_effect=lambda x: x,
    )
    def setUpClass(cls, m):
        cls.patched_fix_dst = m
        directory_of_this_file = os.path.dirname(os.path.abspath(__file__))
        meteologger_storage = MeteologgerStorage_wdat5(
            {
                "station_id": 1334,
                "path": os.path.join(directory_of_this_file, "wdat5_data"),
                "storage_format": "wdat5",
                "outsidetemp": 1256,
                "hioutsidetemp": 1257,
                "rain": 1652,
            }
        )
        cls.result = meteologger_storage._get_storage_tail(
            dt.datetime(2013, 12, 24, 22, 30)
        )

    def test_length(self):
        # The January 2014 file goes as far as 2014-01-03 12:10. Therefore we have 9
        # whole days (from 25 Dec to 2 Jan) or 9*144 = 1296 ten-minute intervals.
        # Plus 8 intervals from 2013-12-24 22:40 to 2013-12-24 23:50, plus 74 intervals
        # from 2014-01-03 00:00 to 12:10. Total 1378.
        self.assertEqual(len(self.result), 1378)

    def test_first_timestamp(self):
        self.assertEqual(self.result[0]["timestamp"], dt.datetime(2013, 12, 24, 22, 40))

    def test_last_timestamp(self):
        self.assertEqual(self.result[-1]["timestamp"], dt.datetime(2014, 1, 3, 12, 10))

    def test_first_outside_temp(self):
        self.assertAlmostEqual(self.result[0]["outsidetemp"], 8.7, places=1)

    def test_last_outside_temp(self):
        self.assertAlmostEqual(self.result[-1]["outsidetemp"], 13.6, places=1)

    def test_first_hum(self):
        self.assertEqual(self.result[0]["outsidehum"], 88)

    def test_last_hum(self):
        self.assertEqual(self.result[-1]["outsidehum"], 72)

    def test_fix_dst_was_called_enough_times(self):
        # ._fix_dst should have been called once for each December timestamp (31*144)
        # plus two whole days in January (2*144) plus 73 intervals from 2014-01-03 00:10
        # to 2014-01-03 12:10, total 4825.
        self.assertEqual(len(self.patched_fix_dst.mock_calls), 4825)

    def test_fix_dst_was_called_starting_with_first_date(self):
        self.assertEqual(
            self.patched_fix_dst.mock_calls[0], call(dt.datetime(2013, 12, 1, 0, 10))
        )

    def test_fix_dst_was_called_ending_with_last_date(self):
        self.assertEqual(
            self.patched_fix_dst.mock_calls[-1], call(dt.datetime(2014, 1, 3, 12, 10))
        )



import datetime as dt
from unittest import TestCase
from unittest.mock import call, patch

from loggertodb.exceptions import ConfigurationError
from loggertodb.meteologgerstorage import MeteologgerStorage_odbc


class ConfigurationTestCase(TestCase):
    def test_raises_error_on_table_missing(self):
        expected_error_message = 'Parameter "table" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            MeteologgerStorage_odbc(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "odbc",
                    "fields": "5, 6",
                    "date_sql": "irrelevant",
                    "data_columns": "irrelevant",
                }
            )

    def test_raises_error_on_date_sql_missing(self):
        expected_error_message = 'Parameter "date_sql" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            MeteologgerStorage_odbc(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "odbc",
                    "fields": "5, 6",
                    "table": "irrelevant",
                    "data_columns": "irrelevant",
                }
            )

    def test_raises_error_on_data_columns_missing(self):
        expected_error_message = 'Parameter "data_columns" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            MeteologgerStorage_odbc(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "odbc",
                    "fields": "5, 6",
                    "table": "irrelevant",
                    "date_sql": "irrelevant",
                }
            )

    def test_accepts_optional_parameters(sef):
        MeteologgerStorage_odbc(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "odbc",
                "fields": "5, 6",
                "table": "irrelevant",
                "date_sql": "irrelevant",
                "data_columns": "irrelevant",
                "date_format": "irrelevant",
                "decimal_separator": "irrelevant",
            }
        )


class GetStorageTailTestCase(TestCase):
    @patch("loggertodb.meteologgerstorage.pyodbc")
    @patch("loggertodb.meteologgerstorage.MeteologgerStorage._fix_dst")
    def setUp(self, m2, m1):
        # Configure pyodbc mock
        self.pyodbc = m1
        attrs = {
            "connect.return_value.cursor.return_value.__iter__.return_value": [
                ["2019-03-02 05:00;43.1;25.2"],
                ["2019-03-02 04:00;42.1;24.2"],
            ]
        }
        self.pyodbc.configure_mock(**attrs)

        # Configure _fix_dst mock
        self.patched_fix_dst = m2
        self.patched_fix_dst.configure_mock(side_effect=lambda x: x)

        self.meteologger_storage = MeteologgerStorage_odbc(
            {
                "station_id": 1334,
                "path": "Some ODBC path",
                "storage_format": "odbc",
                "fields": "5, 6",
                "table": "SomeSQLTable",
                "data_columns": "variable1,variable2",
                "date_format": "%Y-%m-%d %H:%M",
                "date_sql": "timestamp",
                "decimal_separator": ".",
            }
        )
        self.result = self.meteologger_storage._get_storage_tail(
            dt.datetime(2019, 3, 2, 3, 0)
        )

    def test_connected(self):
        self.pyodbc.connect.assert_called_once_with("Some ODBC path")

    def test_executed_sql(self):
        mock_execute = self.pyodbc.connect.return_value.cursor.return_value.execute
        mock_execute.assert_called_once_with(
            """SELECT timestamp + ';' + "variable1" + ';' + "variable2" """
            """FROM "SomeSQLTable" ORDER BY -id"""
        )

    def test_result_length(self):
        self.assertEqual(len(self.result), 2)

    def test_result_timestamp_1(self):
        self.assertEqual(self.result[0]["timestamp"], dt.datetime(2019, 3, 2, 4, 0))

    def test_result_line_1(self):
        self.assertEqual(self.result[0]["line"], "2019-03-02 04:00;42.1;24.2")

    def test_result_timestamp_2(self):
        self.assertEqual(self.result[1]["timestamp"], dt.datetime(2019, 3, 2, 5, 0))

    def test_result_line_2(self):
        self.assertEqual(self.result[1]["line"], "2019-03-02 05:00;43.1;25.2")

    def test_called_fix_dst(self):
        self.patched_fix_dst.assert_has_calls(
            [call(dt.datetime(2019, 3, 2, 5, 0)), call(dt.datetime(2019, 3, 2, 4, 0))]
        )



import datetime as dt
import logging
import math
from unittest import TestCase

from loggertodb.exceptions import ConfigurationError, MeteologgerStorageReadError
from loggertodb.meteologgerstorage import MeteologgerStorage_CR1000


class CheckParametersTestCase(TestCase):
    def test_raises_error_on_subset_identifiers_missing(self):
        expected_error_message = 'Parameter "subset_identifiers" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            MeteologgerStorage_CR1000(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "dummy",
                    "fields": "5, 6",
                }
            )


class ExtractTimestampTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_CR1000(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "CR1000",
                "fields": "5, 6",
                "null": "NULL",
                "subset_identifiers": "18",
            },
            logger=dummy_logger,
        )

    def test_extracts_timestamp(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                "2019-02-28 13:47,111,222,18,25.2,42.3\n"
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )

    def test_raises_error_on_invalid_date(self):
        with self.assertRaises(MeteologgerStorageReadError):
            self.meteologger_storage._extract_timestamp(
                "2019-02-29 13:47,111,222,18,25.2,42.3\n"
            ),


class GetItemFromLineTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_CR1000(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "CR1000",
                "fields": "5, 6",
                "null": "NULL",
                "subset_identifiers": "18",
            },
            logger=dummy_logger,
        )

    def test_get_first_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28 13:47,111,222,18,25.2,42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "")

    def test_get_second_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28 13:47,111,222,18,25.2,42.3\n", 2
        )
        self.assertAlmostEqual(r[0], 42.3)
        self.assertEqual(r[1], "")

    def test_get_null_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28 13:47,111,222,18,NULL,42.3\n", 1
        )
        self.assertTrue(math.isnan(r[0]))
        self.assertEqual(r[1], "")

    def test_raises_error_on_invalid_number(self):
        with self.assertRaises(ValueError):
            self.meteologger_storage._get_item_from_line(
                "2019-02-28 13:47,111,222,18,hello,42.3\n", 1
            )


class SubsetIdentifiersMatchTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_CR1000(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "CR1000",
                "fields": "5, 6",
                "null": "NULL",
                "subset_identifiers": "18",
            },
            logger=dummy_logger,
        )

    def test_matches(self):
        self.assertTrue(
            self.meteologger_storage._subset_identifiers_match(
                "2019-02-28 13:47,111,222,18,hello,42.3\n"
            )
        )

    def test_does_not_match(self):
        self.assertFalse(
            self.meteologger_storage._subset_identifiers_match(
                "2019-02-28 13:47,111,222,19,hello,42.3\n"
            )
        )



import datetime as dt
import logging
import math
from unittest import TestCase

from loggertodb.exceptions import MeteologgerStorageReadError
from loggertodb.meteologgerstorage import MeteologgerStorage_deltacom


class ExtractTimestampTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_deltacom(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "deltacom",
                "fields": "5, 6",
                "null": "NULL",
            },
            logger=dummy_logger,
        )

    def test_extracts_timestamp(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp("2019-02-28T13:47 25.2 42.3\n"),
            dt.datetime(2019, 2, 28, 13, 47),
        )

    def test_raises_error_on_invalid_date(self):
        with self.assertRaises(MeteologgerStorageReadError):
            self.meteologger_storage._extract_timestamp("2019-02-29T13:47 25.2 42.3\n"),


class GetItemFromLineTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_deltacom(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "deltacom",
                "fields": "5, 6",
                "null": "NULL",
            },
            logger=dummy_logger,
        )

    def test_get_first_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2 42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "")

    def test_get_second_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2 42.3\n", 2
        )
        self.assertAlmostEqual(r[0], 42.3)
        self.assertEqual(r[1], "")

    def test_get_item_with_overrun(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2# 42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "LOGOVERRUN")

    def test_get_item_with_noisy(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2$ 42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "LOGNOISY")

    def test_get_item_with_outside(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2% 42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "LOGOUTSIDE")

    def test_get_item_with_range(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 25.2& 42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "LOGRANGE")

    def test_raises_error_on_invalid_flag(self):
        with self.assertRaises(ValueError):
            self.meteologger_storage._get_item_from_line(
                "2019-02-28T13:47 25.2! 42.3\n", 1
            )

    def test_get_null_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "2019-02-28T13:47 NULL 42.3\n", 1
        )
        self.assertTrue(math.isnan(r[0]))
        self.assertEqual(r[1], "")



import datetime as dt
import logging
import math
from unittest import TestCase

from loggertodb.exceptions import MeteologgerStorageReadError
from loggertodb.meteologgerstorage import MeteologgerStorage_simple


class CheckParametersTestCase(TestCase):
    def test_accepts_allowed_optional_parameters(self):
        MeteologgerStorage_simple(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "simple",
                "fields": "5, 6",
                "null": "NULL",
                "delimiter": ",",
                "date_format": "%Y-%m-%d %H:%M",
                "nfields_to_ignore": 2,
            }
        )


class ExtractTimestampTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_simple(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "simple",
                "fields": "5, 6",
                "null": "NULL",
                "delimiter": ",",
                "date_format": "%d/%m/%Y %H:%M",
                "nfields_to_ignore": 2,
            },
            logger=dummy_logger,
        )

    def test_extracts_timestamp_with_time_included(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                "ign1,ign2,28/2/2019 13:47,25.2,42.3\n"
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_extracts_timestamp_separate_time(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                "ign1,ign2,28/2/2019,13:47,25.2,42.3\n"
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertTrue(self.meteologger_storage._separate_time)

    def test_raises_error_on_invalid_date(self):
        with self.assertRaises(MeteologgerStorageReadError):
            self.meteologger_storage._extract_timestamp(
                "ign1,ign2,29/2/2019 13:47,25.2,42.3\n"
            )

    def test_ignores_leading_spaces(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                "ign1,ign2, 28/2/2019 13:47,25.2,42.3\n"
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_trailing_spaces(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                "ign1,ign2,28/2/2019 13:47 ,25.2,42.3\n"
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_double_quotes(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                'ign1,ign2,"28/2/2019 13:47",25.2,42.3\n'
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_leading_spaces_before_opening_quote(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                'ign1,ign2, "28/2/2019 13:47",25.2,42.3\n'
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_trailing_spaces_after_closing_quote(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                'ign1,ign2,"28/2/2019 13:47" ,25.2,42.3\n'
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_leading_spaces_after_opening_quote(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                'ign1,ign2," 28/2/2019 13:47",25.2,42.3\n'
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)

    def test_ignores_trailing_spaces_before_closing_quote(self):
        self.assertEqual(
            self.meteologger_storage._extract_timestamp(
                'ign1,ign2,"28/2/2019 13:47 ",25.2,42.3\n'
            ),
            dt.datetime(2019, 2, 28, 13, 47),
        )
        self.assertFalse(self.meteologger_storage._separate_time)


class GetItemFromLineTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_simple(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "simple",
                "fields": "5, 6",
                "null": "NULL",
                "delimiter": ",",
                "date_format": "%d/%m/%Y %H:%M",
                "nfields_to_ignore": 2,
            },
            logger=dummy_logger,
        )
        self.meteologger_storage._separate_time = False

    def test_get_first_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1,ign2,28/2/2019 13:47,25.2,42.3\n", 1
        )
        self.assertAlmostEqual(r[0], 25.2)
        self.assertEqual(r[1], "")

    def test_get_second_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1,ign2,28/2/2019 13:47,25.2,42.3\n", 2
        )
        self.assertAlmostEqual(r[0], 42.3)
        self.assertEqual(r[1], "")

    def test_get_null_item(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1,ign2,28/2/2019 13:47,NULL,42.3\n", 1
        )
        self.assertTrue(math.isnan(r[0]))
        self.assertEqual(r[1], "")

    def test_raises_error_on_invalid_number(self):
        with self.assertRaises(ValueError):
            self.meteologger_storage._get_item_from_line(
                "ign1,ign2,28/2/2019 13:47,hello,42.3\n", 1
            )


class NullTestCase(TestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_simple(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "simple",
                "fields": "5, 6",
                "null": "-6999",
                "delimiter": ",",
                "date_format": "%d/%m/%Y %H:%M",
                "nfields_to_ignore": 2,
            },
            logger=dummy_logger,
        )
        self.meteologger_storage._separate_time = False

    def test_number_smaller_by_more_than_1e6_is_interpreted_as_number(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1, ign2, 28/2/2019 13:47,-6999.000002,42.3\n", 1
        )
        self.assertAlmostEqual(r[0], -6999.000002)

    def test_number_greater_by_more_than_1e6_is_interpreted_as_number(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1, ign2, 28/2/2019 13:47,-6998.999998,42.3\n", 1
        )
        self.assertAlmostEqual(r[0], -6998.999998)

    def test_number_smaller_by_less_than_1e6_is_interpreted_as_null(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1, ign2, 28/2/2019 13:47,-6999.0000007,42.3\n", 1
        )
        self.assertTrue(math.isnan(r[0]))

    def test_number_greater_by_less_than_1e6_is_interpreted_as_null(self):
        r = self.meteologger_storage._get_item_from_line(
            "ign1, ign2, 28/2/2019 13:47,-6998.9999992,42.3\n", 1
        )
        self.assertTrue(math.isnan(r[0]))


class NullStrTestCase(NullTestCase):
    def setUp(self):
        dummy_logger = logging.getLogger("dummy")
        dummy_logger.addHandler(logging.NullHandler())
        self.meteologger_storage = MeteologgerStorage_simple(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "simple",
                "fields": "5, 6",
                "nullstr": "-6999",
                "delimiter": ",",
                "date_format": "%d/%m/%Y %H:%M",
                "nfields_to_ignore": 2,
            },
            logger=dummy_logger,
        )
        self.meteologger_storage._separate_time = False



import os
import textwrap
from unittest import TestCase
from unittest.mock import patch

from click.testing import CliRunner

from loggertodb import cli


class NonExistentConfigFileTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        self.result = runner.invoke(cli.main, ["nonexistent.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "No such file or directory: 'nonexistent.conf'", self.result.stderr
        )


class MissingBaseUrlTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        auth_token = 123456789abcdef0123456789abcdef012345678
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("No option 'base_url' in section: 'General'", self.result.stderr)


class NonExistentLogLevelTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        auth_token = 123456789abcdef0123456789abcdef012345678
                        loglevel = NONEXISTENT_LOG_LEVEL
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "loglevel must be one of ERROR, WARNING, INFO, DEBUG", self.result.stderr
        )


class ConfigurationWithNoMeteologgersTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        auth_token = 123456789abcdef0123456789abcdef012345678
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("No stations have been specified", self.result.stderr)


class ConfigurationWithUnsupportedFormatTestCase(TestCase):
    @patch("loggertodb.cli.Enhydris")
    def setUp(self, mock_enhydris):
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        auth_token = 123456789abcdef0123456789abcdef012345678

                        [My station]
                        storage_format = unsupported
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn("Unsupported format 'unsupported'", self.result.stderr)


class CorrectConfigurationTestCase(TestCase):
    @patch("loggertodb.cli.Enhydris")
    @patch("loggertodb.meteologgerstorage.MeteologgerStorage_simple")
    def setUp(self, mock_meteologgerstorage, mock_enhydris):
        self.mock_meteologgerstorage = mock_meteologgerstorage
        self.mock_enhydris = mock_enhydris
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        auth_token = 123456789abcdef0123456789abcdef012345678

                        [My station]
                        storage_format = simple
                        station_id = 1334
                        path = .
                        fields = 1,2,3
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 0)

    def test_has_used_base_url(self):
        configuration = self.mock_enhydris.call_args[0][0]
        self.assertEqual(configuration.base_url, "https://example.com")

    def test_has_used_auth_token(self):
        configuration = self.mock_enhydris.call_args[0][0]
        self.assertEqual(
            configuration.auth_token, "123456789abcdef0123456789abcdef012345678"
        )

    def test_has_uploaded(self):
        self.mock_enhydris.return_value.upload.assert_called_once_with(
            self.mock_meteologgerstorage.return_value
        )


class CorrectConfigurationWithLogFileTestCase(TestCase):
    @patch("loggertodb.cli.Enhydris")
    @patch("loggertodb.meteologgerstorage.MeteologgerStorage_simple")
    def test_creates_log_file(self, *args):
        self.mock_meteologgerstorage = args[0]
        self.mock_enhydris = args[1]
        runner = CliRunner(mix_stderr=False)
        with runner.isolated_filesystem():
            with open("loggertodb.conf", "w") as f:
                f.write(
                    textwrap.dedent(
                        """\
                        [General]
                        base_url = https://example.com
                        auth_token = 123456789abcdef0123456789abcdef012345678
                        logfile = deleteme

                        [My station]
                        storage_format = simple
                        station_id = 1334
                        path = .
                        fields = 1,2,3
                        """
                    )
                )
            self.result = runner.invoke(cli.main, ["loggertodb.conf"])
            self.assertTrue(os.path.exists("deleteme"))



import datetime as dt
from collections import OrderedDict
from unittest import TestCase
from unittest.mock import patch

import pandas as pd

from loggertodb import ConfigurationError
from loggertodb.meteologgerstorage import MeteologgerStorage


class DummyMeteologgerStorage(MeteologgerStorage):
    def _get_storage_tail(self, after_timestamp):
        return [
            {"timestamp": after_timestamp + dt.timedelta(minutes=1), "line": "line1"},
            {"timestamp": after_timestamp + dt.timedelta(minutes=2), "line": "line2"},
        ]

    def _extract_value_and_flags(self, ts_id, record):
        # Use the ts_id as value and the whole line as flags
        return (ts_id, record["line"])

    @property
    def timeseries_group_ids(self):
        return (15, 16)


class DummyWrongOrderMeteologgerStorage(DummyMeteologgerStorage):
    """A file with a storage tail that contains records in the wrong order."""

    def _get_storage_tail(self, after_timestamp):
        result = super()._get_storage_tail(after_timestamp)
        result.reverse()
        return result


class MeteologgerStorageCheckParametersTestCase(TestCase):
    def test_raises_error_on_path_missing(self):
        expected_error_message = 'Parameter "path" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyMeteologgerStorage({"station_id": 1334, "storage_format": "dummy"})

    def test_raises_error_on_storage_format_missing(self):
        expected_error_message = 'Parameter "storage_format" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyMeteologgerStorage({"station_id": 1334, "path": "irrelevant"})

    def test_raises_error_on_station_missing(self):
        expected_error_message = 'Parameter "station_id" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyMeteologgerStorage({"path": "irrelevant", "storage_format": "dummy"})

    def test_raises_error_on_invalid_parameter(self):
        expected_error_message = 'Unknown parameter "hello"'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyMeteologgerStorage(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "dummy",
                    "hello": "world",
                }
            )


class MeteologgerStorageGetRecentDataTestCase(TestCase):
    def setUp(self):
        self.storage = DummyMeteologgerStorage(
            {"station_id": 1334, "path": "irrelevant", "storage_format": "dummy"}
        )
        self.result = self.storage.get_recent_data(15, dt.datetime(2019, 2, 27, 12, 52))

    def test_return_value(self):
        expected_result = pd.DataFrame(
            data=OrderedDict([("value", [15, 15]), ("flags", ["line1", "line2"])]),
            index=[dt.datetime(2019, 2, 27, 12, 53), dt.datetime(2019, 2, 27, 12, 54)],
            dtype=object,
        )
        pd.testing.assert_frame_equal(self.result, expected_result)

    def test_cached_return_value(self):
        # We have already called get_recent_data(). Now we call it again, for another
        # timeseries, and a timestamp later than the previous one. It should work
        # correctly.
        second_result = self.storage.get_recent_data(
            16, dt.datetime(2019, 2, 27, 12, 53)
        )
        expected_result = pd.DataFrame(
            data=OrderedDict([("value", [16]), ("flags", ["line2"])]),
            index=[dt.datetime(2019, 2, 27, 12, 54)],
            dtype=object,
        )
        pd.testing.assert_frame_equal(second_result, expected_result)


class MeteologgerStorageGetRecentDataWrongOrderTestCase(TestCase):
    def setUp(self):
        self.storage = DummyWrongOrderMeteologgerStorage(
            {"station_id": 1334, "path": "irrelevant", "storage_format": "dummy"}
        )

    def test_exception(self):
        msg = "incorrectly ordered after 2019-02-27T12:54:00"
        with self.assertRaisesRegex(ValueError, msg):
            self.result = self.storage.get_recent_data(
                15, dt.datetime(2019, 2, 27, 12, 52)
            )


class PatchableDatetime(dt.datetime):
    """Patchable version of dt.datetime.

    mock.patch() can't patch dt.datetime.now(), because dt.datetime is written in C. We
    therefore create this subclass. Whenever we need to patch dt.datetime.now(), we
    first patch dt.datetime with this subclass, and then we can patch dt.datetime.now().
    """

    pass


class EETTimezone(dt.tzinfo):
    def utcoffset(self, adatetime):
        return dt.timedelta(hours=2)

    def dst(self, adatetime):
        return None

    def tzname(self, adatetime):
        return "EET"


class EESTTimezone(dt.tzinfo):
    def utcoffset(self, adatetime):
        return dt.timedelta(hours=3)

    def dst(self, adatetime):
        return dt.timedelta(hours=1)

    def tzname(self, adatetime):
        return "EEST"


class MeteologgerStorageFixDstTestCase(TestCase):
    def setUp(self):
        self.storage = DummyMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "dummy",
                "timezone": "Europe/Athens",
            }
        )

    def test_date_without_dst_is_returned_as_is(self):
        self.assertEqual(
            self.storage._fix_dst(dt.datetime(2019, 2, 27, 13, 39)),
            dt.datetime(2019, 2, 27, 13, 39),
        )

    def test_date_with_dst_is_returned_with_dst_removed(self):
        self.assertEqual(
            self.storage._fix_dst(dt.datetime(2019, 4, 27, 13, 39)),
            dt.datetime(2019, 4, 27, 12, 39),
        )

    @patch(
        "loggertodb.meteologgerstorage.dt.datetime.now",
        return_value=dt.datetime(2018, 10, 28, 5, 0, tzinfo=EETTimezone()),
    )
    @patch("loggertodb.meteologgerstorage.dt.datetime", new=PatchableDatetime)
    def test_ambiguous_date_is_returned_as_is_after_dst_switch(self, m):
        self.assertEqual(
            self.storage._fix_dst(dt.datetime(2018, 10, 28, 3, 30)),
            dt.datetime(2018, 10, 28, 3, 30),
        )

    @patch(
        "loggertodb.meteologgerstorage.dt.datetime.now",
        return_value=dt.datetime(2018, 10, 28, 1, 0, tzinfo=EESTTimezone()),
    )
    @patch("loggertodb.meteologgerstorage.dt.datetime", new=PatchableDatetime)
    def test_ambiguous_date_is_returned_with_dst_removed_before_dst_switch(self, m):
        self.assertEqual(
            self.storage._fix_dst(dt.datetime(2018, 10, 28, 3, 30)),
            dt.datetime(2018, 10, 28, 2, 30),
        )



import configparser
import datetime as dt
import logging
import sys
import traceback

import click

from . import __version__, meteologgerstorage
from .enhydris import Enhydris
from .exceptions import LoggerToDbError
from .upgrade import ConfigFile


class WrongValueError(configparser.Error):
    pass


class UnsupportedFormat(Exception):
    def __str__(self):
        return "Unsupported format '{}'".format(self.args[0])


@click.command()
@click.option(
    "--upgrade", is_flag=True, default=False, help="Upgrade configuration file"
)
@click.argument("configfile")
@click.version_option(version=__version__, prog_name="loggertodb")
def main(upgrade, configfile):
    """Insert meteorological logger data to Enhydris"""
    if upgrade:
        ConfigFile(configfile).upgrade()
    else:
        LoggerToDb(configfile).run()


class LoggerToDb:
    def __init__(self, configfile):
        self.configfile = configfile
        self.logging_system = Logging()

    def run(self):
        try:
            self.configuration = Configuration(self.configfile, self.logging_system)
            self.configuration.read()
            self.logging_system.setup_logger(self.configuration)
            self.logging_system.log_start_of_execution()
            self.enhydris = Enhydris(self.configuration)
            self._process_stations()
            self.logging_system.log_end_of_execution()
        except Exception as e:
            self.logging_system.logger.error(str(e))
            self.logging_system.logger.debug(traceback.format_exc())
            raise click.ClickException(str(e))

    def _process_stations(self):
        for meteologger_storage in self.configuration.meteologger_storages:
            try:
                self.enhydris.upload(meteologger_storage)
            except LoggerToDbError as e:
                msg = "Error while processing item {}: {}".format(
                    meteologger_storage.section, str(e)
                )
                sys.stderr.write(msg + "\n")
                self.logging_system.logger.error(msg)
                self.logging_system.logger.debug(traceback.format_exc())


class Logging:
    def __init__(self):
        self.logger = logging.getLogger("loggertodb")
        self.stdout_handler = logging.StreamHandler()
        self.logger.addHandler(self.stdout_handler)

    def setup_logger(self, configuration):
        self.logger.setLevel(configuration.loglevel.upper())
        if configuration.logfile:
            self.logger.removeHandler(self.stdout_handler)
            self.logger.addHandler(logging.FileHandler(configuration.logfile))

    def log_start_of_execution(self):
        self.logger.info("Starting loggertodb, " + dt.datetime.today().isoformat())

    def log_end_of_execution(self):
        self.logger.info("Finished loggertodb, " + dt.datetime.today().isoformat())


class Configuration:
    def __init__(self, configfile, logging_system):
        self.logging_system = logging_system
        self.configfile = configfile
        self.config = configparser.ConfigParser(interpolation=None)
        with open(self.configfile) as f:
            self.config.read_file(f)
        self.meteologger_storages = []

    def read(self):
        self._read_general_section()
        self._read_station_sections()

    def _read_general_section(self):
        self.base_url = self.config.get("General", "base_url")
        self.auth_token = self.config.get("General", "auth_token")
        self.logfile = self.config.get("General", "logfile", fallback="")
        self.loglevel = self.config.get("General", "loglevel", fallback="warning")
        log_levels = ("ERROR", "WARNING", "INFO", "DEBUG")
        if self.loglevel.upper() not in log_levels:
            raise WrongValueError("loglevel must be one of " + ", ".join(log_levels))

    def _read_station_sections(self):
        station_section_names = [n for n in self.config.sections() if n != "General"]
        if not len(station_section_names):
            raise configparser.NoSectionError("No stations have been specified")
        for section_name in station_section_names:
            section = self.config[section_name]
            klassname = "MeteologgerStorage_" + section["storage_format"]
            if not hasattr(meteologgerstorage, klassname):
                raise UnsupportedFormat(section["storage_format"])
            klass = getattr(meteologgerstorage, klassname)
            self.meteologger_storages.append(
                klass(section, logger=self.logging_system.logger)
            )


if __name__ == "__main__":
    sys.exit(main())



import datetime as dt
from collections import namedtuple

from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries

CompositeTimeseriesId = namedtuple(
    "CompositeTimeseriesId", ("timeseries_group_id", "timeseries_id")
)


class Enhydris:
    def __init__(self, configuration):
        self.base_url = configuration.base_url
        self.auth_token = configuration.auth_token
        self.client = EnhydrisApiClient(self.base_url, self.auth_token)

    def upload(self, meteologger_storage):
        self._meteologger_storage = meteologger_storage
        self._get_composite_timeseries_ids()
        self._get_ts_end_dates()
        self._upload_all_new_data()

    def _get_composite_timeseries_ids(self):
        """Create a list of (timeseries_group_id, initial_timeseries_id) pairs."""
        station_id = self._meteologger_storage.station_id
        self._composite_timeseries_ids = []
        for timeseries_group_id in self._meteologger_storage.timeseries_group_ids:
            timeseries_id = self._get_timeseries_id(station_id, timeseries_group_id)
            self._composite_timeseries_ids.append(
                CompositeTimeseriesId(timeseries_group_id, timeseries_id)
            )

    def _get_timeseries_id(self, station_id, timeseries_group_id):
        timeseries = self.client.list_timeseries(station_id, timeseries_group_id)
        for item in timeseries:
            if item["type"] == "Initial":
                return item["id"]
        return self._create_timeseries(station_id, timeseries_group_id)

    def _create_timeseries(self, station_id, timeseries_group_id):
        return self.client.post_timeseries(
            station_id,
            timeseries_group_id,
            data={
                "type": "Initial",
                "time_step": "",
                "timeseries_group": timeseries_group_id,
            },
        )

    def _get_ts_end_dates(self):
        station_id = self._meteologger_storage.station_id
        start_of_time = dt.datetime(1700, 1, 1)
        self._ts_end_dates = {
            cts_id: self.client.get_ts_end_date(station_id, *cts_id) or start_of_time
            for cts_id in self._composite_timeseries_ids
        }

    def _upload_all_new_data(self):
        station_id = self._meteologger_storage.station_id
        sorted_ts_end_dates = sorted(self._ts_end_dates.items(), key=lambda x: x[1])
        for cts_id, ts_end_date in sorted_ts_end_dates:
            new_data = self._meteologger_storage.get_recent_data(
                cts_id.timeseries_group_id, ts_end_date
            )
            if len(new_data):
                self.client.post_tsdata(station_id, *cts_id, HTimeseries(new_data))



import datetime as dt
import logging
import os
import re
import struct
from abc import ABC, abstractmethod
from glob import glob

import iso8601
import numpy as np
import pandas as pd
from pytz import timezone
from simpletail import ropen

from .exceptions import ConfigurationError, MeteologgerStorageReadError

pyodbc = None


class MeteologgerStorage(ABC):
    def __init__(self, parameters, logger=None):
        self._reset_ambiguous_hour_data()
        self.__check_parameters(parameters)
        self.station_id = int(parameters["station_id"])
        self.path = parameters["path"]
        self.timezone = timezone(parameters.get("timezone", "UTC"))
        self.logger = logger
        if not self.logger:
            self.logger = logging.getLogger("loggerstorage")
            self.logger.setLevel(logging.WARNING)
            self.logger.addHandler(logging.StreamHandler())

    def __check_parameters(self, parameters):
        # Check that all required parameters are present
        for parameter in self.get_required_parameters():
            if parameter not in parameters:
                raise ConfigurationError('Parameter "{}" is required'.format(parameter))

        # Check that all given parameters are valid
        all_parameters = self.get_required_parameters() | self.get_optional_parameters()
        for parameter in parameters:
            if parameter not in all_parameters:
                raise ConfigurationError('Unknown parameter "{}"'.format(parameter))

    @property
    @abstractmethod
    def timeseries_group_ids(self):
        pass

    def get_recent_data(self, ts_group_id, after_timestamp):
        self._reset_ambiguous_hour_data()
        cached_after_timestamp = getattr(
            self, "_cached_after_timestamp", dt.datetime(9999, 12, 31)
        )
        if after_timestamp < cached_after_timestamp:
            self._extract_data(after_timestamp=after_timestamp)
        from_timestamp = after_timestamp + dt.timedelta(seconds=1)
        self._check_monotonic(self._cached_data[ts_group_id].index)
        return self._cached_data[ts_group_id].loc[from_timestamp:]

    def _check_monotonic(self, index):
        if index.is_monotonic:
            return
        else:
            self._raise_monotonic_exception(index)

    def _raise_monotonic_exception(self, index):
        offending_date = self._locate_first_nonmonotonic_date(index)
        raise ValueError(
            "Data is incorrectly ordered after " + offending_date.isoformat()
        )

    def _locate_first_nonmonotonic_date(self, index):
        prev = None
        for current in index:
            if prev is not None and prev > current:
                return prev
            prev = current

    def _extract_data(self, after_timestamp):
        """Extract the part of the storage that is after_timestamp.

        Reads the part of the storage that is after_timestamp and puts it in
        self._cached_data, a dictionary mapping time series ids to pandas dataframes
        with the data.

        Also sets self._cached_after_timestamp.
        """

        # Read part of storage later than after_timestamp
        self.logger.info(
            "Reading data storage newer than {}".format(after_timestamp.isoformat())
        )
        storage_tail = self._get_storage_tail(after_timestamp)
        self.logger.info("{} new lines".format(len(storage_tail)))
        if len(storage_tail) > 0:
            self.logger.info(
                "First new date: {}".format(storage_tail[0]["timestamp"].isoformat())
            )

        # Start with empty time series
        index = np.empty(len(storage_tail), dtype="datetime64[s]")
        data = {
            ts_id: np.empty((len(storage_tail), 2), dtype=object)
            for ts_id in self.timeseries_group_ids
        }

        # Iterate through the storage tail and fill in the time series
        try:
            for i, record in enumerate(storage_tail):
                for ts_id in self.timeseries_group_ids:
                    v, f = self._extract_value_and_flags(ts_id, record)
                    index[i] = np.datetime64(record["timestamp"])
                    data[ts_id][i, 0] = v
                    data[ts_id][i, 1] = f
        except ValueError as e:
            message = "parsing error while trying to read values: " + str(e)
            self._raise_error(record["line"], message)

        # Replace self._cached_data and self._after_timestamp, if any
        self._cached_data = {
            tsg_id: pd.DataFrame(
                columns=["value", "flags"], index=index, data=data[tsg_id]
            )
            for tsg_id in self.timeseries_group_ids
        }
        self._cached_after_timestamp = after_timestamp

    @abstractmethod
    def _get_storage_tail(self, after_timestamp):
        """Read the part of the data storage after_timestamp.

        Returns a list of dictionaries. Each of these dictionaries is a measurement
        record (typically with many measurements).  The first item in the returned list
        is the first record that has a timestamp later than after_timestamp.

        Each dictionary has a "timestamp" key, a datetime object. The rest of the keys
        depend on the specific storage class. For TextFileMeteologgerStorage, there's
        one more key, "line", with a string with the entire line in the file. For wdat5,
        there are keys like "outsidetemp" etc., holding the values of the variables.
        """

    def _fix_dst(self, adatetime):
        """Remove any DST from a date.

        Determine if a date contains DST. If it does, remove the extra hour. Returns the
        fixed date.

        If adatetime is an ambiguous hour, then we try to deduce whether it is in dst or
        not. If the switch hasn't occurred yet, we conclude it's in dst. If the switch
        has occurred, we conclude it's not in dst, unless we've already seen that date
        (i.e. already dealt with it in a previous call—remember we process files
        backwards, from later to earlier dates); in that case, we conclude it's in dst.
        """
        if not hasattr(self.timezone, "_utc_transition_times"):
            return adatetime  # Timezone does not switch to DST
        is_dst = self._auto_detect_is_dst(adatetime)
        return adatetime - self.timezone.dst(adatetime, is_dst=is_dst)

    def _reset_ambiguous_hour_data(self):
        self._dates_already_seen = set()
        self._we_are_in_the_second_occurrence_of_the_ambigous_hour = False

    def _auto_detect_is_dst(self, adatetime):
        if self._datetime_is_ambiguous(adatetime):
            return self._determine_is_dst_for_ambiguous_hour(adatetime)
        else:
            self._reset_ambiguous_hour_data()
            return None

    def _datetime_is_ambiguous(self, adatetime):
        dst = self.timezone.dst
        return dst(adatetime, is_dst=True) != dst(adatetime, is_dst=False)

    def _determine_is_dst_for_ambiguous_hour(self, adatetime):
        if self._switch_has_not_occurred(adatetime):
            return True
        if self._we_are_in_the_second_occurrence_of_the_ambigous_hour:
            return True
        if adatetime in self._dates_already_seen:
            self._we_are_in_the_second_occurrence_of_the_ambigous_hour = True
            return True
        self._dates_already_seen.add(adatetime)
        return False

    def _switch_has_not_occurred(self, adatetime):
        now = dt.datetime.now(self.timezone)
        now_naive = now.replace(tzinfo=None)
        if abs(adatetime - now_naive) > dt.timedelta(hours=24):
            return False
        return bool(now.dst())

    def _raise_error(self, line, msg):
        filename = self.filename if hasattr(self, "filename") else self.path
        errmessage = '{}: "{}": {}'.format(filename, line, msg)
        self.logger.error("Error while parsing, message: " + errmessage)
        raise MeteologgerStorageReadError(errmessage)

    def get_required_parameters(self):
        return {"path", "storage_format", "station_id"}

    def get_optional_parameters(self):
        return {"timezone"}

    @abstractmethod
    def _extract_value_and_flags(self, ts_id, record):
        pass

    def _is_null(self, value):
        if not self.null:
            return False
        try:
            null = float(self.null)
            return abs(float(value) - null) < 1e-6
        except ValueError:
            return value == self.null


class TextFileMeteologgerStorage(MeteologgerStorage):
    def __init__(self, parameters, logger=None):
        super().__init__(parameters, logger=logger)
        self.fields = [int(x) for x in parameters.get("fields", "").split(",") if x]
        self.subset_identifiers = parameters.get("subset_identifiers", "")
        self.delimiter = parameters.get("delimiter", None)
        self.decimal_separator = parameters.get("decimal_separator", "")
        self.date_format = parameters.get("date_format", "")
        self.null = parameters.get("null", parameters.get("nullstr", ""))
        self.nfields_to_ignore = int(parameters.get("nfields_to_ignore", "0"))
        self.ignore_lines = parameters.get("ignore_lines", "")
        self.encoding = parameters.get("encoding", "utf8")

    def get_required_parameters(self):
        return super().get_required_parameters() | {"fields"}

    def get_optional_parameters(self):
        return super().get_optional_parameters() | {
            "null",
            "nullstr",
            "ignore_lines",
            "encoding",
        }

    @property
    def timeseries_group_ids(self):
        return set(self.fields) - {0}

    def _subset_identifiers_match(self, line):
        return True

    @abstractmethod
    def _extract_timestamp(self, line):
        pass

    def _extract_value_and_flags(self, ts_id, record):
        for seq, tid in enumerate(self.fields, start=1):
            if tid == ts_id:
                break
        v, f = self._get_item_from_line(record["line"], seq)
        if self.decimal_separator and (self.decimal_separator != "."):
            v = v.replace(self.decimal_separator, ".")
        return v, f

    @abstractmethod
    def _get_item_from_line(self, line, seq):
        pass

    def _get_storage_tail(self, after_timestamp):
        return self._get_storage_tail_from_file(self.path, after_timestamp)[0]

    def _get_storage_tail_from_file(self, filename, after_timestamp):
        result = []
        reached_after_timestamp = False
        with ropen(filename, encoding=self.encoding, errors="replace") as xr:
            prev_timestamp = ""
            for line in xr:
                if self._must_ignore_line(line):
                    self.logger.debug("Ignoring line '{}'".format(line))
                    continue
                else:
                    self.logger.debug("Parsing line '{}'".format(line))

                timestamp = self._extract_timestamp(line).replace(second=0)
                timestamp = self._fix_dst(timestamp)
                if timestamp == prev_timestamp:
                    w = "Omitting line with repeated timestamp " + timestamp.isoformat()
                    self.logger.warning(w)
                    continue
                prev_timestamp = timestamp
                self.logger.debug("Timestamp: " + timestamp.isoformat())
                if timestamp <= after_timestamp:
                    reached_after_timestamp = True
                    break
                result.append({"timestamp": timestamp, "line": line})

        result.reverse()
        return (result, reached_after_timestamp)

    def _must_ignore_line(self, line):
        if not line.strip() or not self._subset_identifiers_match(line):
            return True
        if not self.ignore_lines:
            return False
        return bool(re.search(self.ignore_lines, line))


class MultiTextFileMeteologgerStorage(TextFileMeteologgerStorage):
    def _get_storage_tail(self, after_timestamp):
        self._get_sorted_files()
        return self._get_storage_tail_from_multiple_files(after_timestamp)

    def _get_sorted_files(self):
        files = []
        for filename in glob(self.path):
            self.filename = filename
            files.append(self._get_file())
        self._sorted_files = sorted(
            files, key=lambda x: x["last_date"] or dt.datetime(1700, 1, 1, 0, 0)
        )

    def _get_file(self):
        return {
            "filename": self.filename,
            "first_date": self._extract_first_date_from_file(self.filename),
            "last_date": self._extract_last_date_from_file(self.filename),
        }

    def _extract_first_date_from_file(self, filename):
        with open(filename, encoding=self.encoding, errors="replace") as f:
            for line in f:
                if self._must_ignore_line(line):
                    continue
                timestamp = self._extract_timestamp(line).replace(second=0)
                timestamp = self._fix_dst(timestamp)
                return timestamp
        return None

    def _extract_last_date_from_file(self, filename):
        with ropen(filename, encoding=self.encoding, errors="replace") as xr:
            for line in xr:
                if self._must_ignore_line(line):
                    continue
                timestamp = self._extract_timestamp(line).replace(second=0)
                timestamp = self._fix_dst(timestamp)
                return timestamp
        return None

    def _get_storage_tail_from_multiple_files(self, after_timestamp):
        result = []
        for file in reversed(self._sorted_files):
            partial_result, reached_after_timestamp = self._get_storage_tail_from_file(
                file["filename"], after_timestamp
            )
            result = partial_result + result
            if reached_after_timestamp:
                break
        return result

    def _raise_monotonic_exception(self, index):
        self._check_file_order()
        super()._raise_monotonic_exception(index)

    def _check_file_order(self):
        previous_file = None
        for file in self._sorted_files:
            self._check_file_dates(file)
            if previous_file:
                self._check_adjacent_file_dates(previous_file, file)
            previous_file = file

    def _check_file_dates(self, file):
        if file["first_date"] is None and file["last_date"] is None:
            return
        assert file["first_date"] is not None and file["last_date"] is not None
        if file["first_date"] <= file["last_date"]:
            return
        raise ValueError(
            "The order of timestamps in file {} is mixed up.".format(file["filename"])
        )

    def _check_adjacent_file_dates(self, file1, file2):
        if file1["last_date"] is None or file2["first_date"] is None:
            return
        if file1["last_date"] < file2["first_date"]:
            return
        raise ValueError(
            "The timestamps in files {} and {} overlap.".format(
                file1["filename"], file2["filename"]
            )
        )


class MeteologgerStorage_deltacom(TextFileMeteologgerStorage):
    deltacom_flags = {
        "#": "LOGOVERRUN",
        "$": "LOGNOISY",
        "%": "LOGOUTSIDE",
        "&": "LOGRANGE",
    }

    def _extract_timestamp(self, line):
        try:
            return iso8601.parse_date(line.split()[0], default_timezone=None)
        except (ValueError, iso8601.ParseError):
            self._raise_error(line, "parse error or invalid date")

    def _get_item_from_line(self, line, seq):
        flags = ""
        item = line.split()[seq].strip()
        if item[-1] in self.deltacom_flags.keys():
            flags = self.deltacom_flags[item[-1]]
            item = item[:-1]
        if self._is_null(item):
            item = "NaN"
        return (float(item), flags)


class MeteologgerStorage_pc208w(TextFileMeteologgerStorage):
    def get_required_parameters(self):
        return super().get_required_parameters() | {"subset_identifiers"}

    def _extract_timestamp(self, line):
        try:
            items = line.split(",")
            year = int(items[2])
            yday = int(items[3])
            hour = int(items[4]) // 100
            minute = int(items[4]) % 100
            if hour == 24:
                hour = 0
                yday = yday + 1
            return dt.datetime(year, 1, 1, hour, minute) + dt.timedelta(yday - 1)
        except (IndexError, ValueError):
            self._raise_error(line, "parse error or invalid date")

    def _get_item_from_line(self, line, seq):
        try:
            item = line.split(",")[seq + 4].strip()
        except IndexError:
            raise ValueError()
        if self._is_null(item):
            item = "NaN"
        return (float(item), "")

    def _subset_identifiers_match(self, line):
        si = line.split(",")[0].strip()
        return si == self.subset_identifiers


class MeteologgerStorage_CR1000(TextFileMeteologgerStorage):
    def get_required_parameters(self):
        return super().get_required_parameters() | {"subset_identifiers"}

    def _extract_timestamp(self, line):
        try:
            datestr = line.split(",")[0].strip('"')
            return iso8601.parse_date(datestr[:16], default_timezone=None)
        except (IndexError, iso8601.ParseError):
            self._raise_error(line, "parse error or invalid date")

    def _get_item_from_line(self, line, seq):
        item = line.split(",")[seq + 3].strip()
        if self._is_null(item):
            item = "NaN"
        return (float(item), "")

    def _subset_identifiers_match(self, line):
        si = line.split(",")[3].strip()
        return si == self.subset_identifiers


class MeteologgerStorage_simple(MultiTextFileMeteologgerStorage):
    def get_optional_parameters(self):
        more_parms = {"nfields_to_ignore", "delimiter", "date_format"}
        return super().get_optional_parameters() | more_parms

    def _extract_timestamp(self, line):
        try:
            items = line.split(self.delimiter)
            datestr = items[self.nfields_to_ignore]
            datestr = datestr.strip().strip('"').strip()
            self._separate_time = False
            if len(datestr) <= 10:
                datestr += " " + items[self.nfields_to_ignore + 1].strip('"')
                self._separate_time = True
            if self.date_format:
                result = dt.datetime.strptime(datestr, self.date_format).replace(
                    second=0
                )
            else:
                result = iso8601.parse_date(datestr[:16], default_timezone=None)
            return result
        except ValueError as e:
            self._raise_error(
                line.strip(), "invalid date '{0}': {1}".format(datestr, str(e))
            )
        except IndexError:
            self._raise_error(line.strip(), "Malformed line")

    def _get_item_from_line(self, line, seq):
        index = self.nfields_to_ignore + seq + (1 if self._separate_time else 0)
        value = line.split(self.delimiter)[index].strip().strip('"').strip()
        if self._is_null(value):
            value = "NaN"
        return (float(value), "")


class MeteologgerStorage_lastem(TextFileMeteologgerStorage):
    def get_required_parameters(self):
        return super().get_required_parameters() | {"subset_identifiers"}

    def get_optional_parameters(self):
        more_parms = {"decimal_separator", "delimiter", "date_format"}
        return super().get_optional_parameters() | more_parms

    def _extract_timestamp(self, line):
        try:
            date = line.split(self.delimiter)[3]
            return dt.datetime.strptime(date, self.date_format)
        except (IndexError, ValueError):
            self._raise_error(line, "parse error or invalid date")

    def _get_item_from_line(self, line, seq):
        value = line.split(self.delimiter)[seq + 3]
        if self._is_null(value):
            value = "NaN"
        else:
            value = value.replace(self.decimal_separator, ".")
        return (float(value), "")

    def _subset_identifiers_match(self, line):
        si = [x.strip() for x in line.split(self.delimiter)[0:3]]
        si1 = [x.strip() for x in self.subset_identifiers.split(",")]
        return si == si1


class MeteologgerStorage_wdat5(MeteologgerStorage):
    wdat_record_format = [
        "<b dataType",
        "<b archiveInterval",
        "<b iconFlags",
        "<b moreFlags",
        "<h packedTime",
        "<h outsideTemp",
        "<h hiOutsideTemp",
        "<h lowOutsideTemp",
        "<h insideTemp",
        "<h barometer",
        "<h outsideHum",
        "<h insideHum",
        "<H rain",
        "<h hiRainRate",
        "<h windSpeed",
        "<h hiWindSpeed",
        "<b windDirection",
        "<b hiWindDirection",
        "<h numWindSamples",
        "<h solarRad",
        "<h hiSolarRad",
        "<B UV",
        "<B hiUV",
        "<b leafTemp1",
        "<b leafTemp2",
        "<b leafTemp3",
        "<b leafTemp4",
        "<h extraRad",
        "<h newSensors1",
        "<h newSensors2",
        "<h newSensors3",
        "<h newSensors4",
        "<h newSensors5",
        "<h newSensors6",
        "<b forecast",
        "<B ET",
        "<b soilTemp1",
        "<b soilTemp2",
        "<b soilTemp3",
        "<b soilTemp4",
        "<b soilTemp5",
        "<b soilTemp6",
        "<b soilMoisture1",
        "<b soilMoisture2",
        "<b soilMoisture3",
        "<b soilMoisture4",
        "<b soilMoisture5",
        "<b soilMoisture6",
        "<b leafWetness1",
        "<b leafWetness2",
        "<b leafWetness3",
        "<b leafWetness4",
        "<b extraTemp1",
        "<b extraTemp2",
        "<b extraTemp3",
        "<b extraTemp4",
        "<b extraTemp5",
        "<b extraTemp6",
        "<b extraTemp7",
        "<b extraHum1",
        "<b extraHum2",
        "<b extraHum3",
        "<b extraHum4",
        "<b extraHum5",
        "<b extraHum6",
        "<b extraHum7",
    ]
    variables_labels = [x.split()[1].lower() for x in wdat_record_format[5:]]

    def get_optional_parameters(self):
        return (
            super().get_optional_parameters()
            | set(self.variables_labels)
            | {
                "temperature_unit",
                "rain_unit",
                "wind_speed_unit",
                "pressure_unit",
                "matric_potential_unit",
            }
        )

    @property
    def timeseries_group_ids(self):
        return {self.variables[lab] for lab in self.variables if self.variables[lab]}

    def __init__(self, parameters, logger=None):
        super().__init__(parameters, logger)

        self.variables = {}
        for label in self.variables_labels:
            self.variables[label] = parameters.get(label)

        unit_parameters = {
            "temperature_unit": ("C", "F"),
            "rain_unit": ("mm", "inch"),
            "wind_speed_unit": ("m/s", "mph"),
            "pressure_unit": ("hPa", "inch Hg"),
            "matric_potential_unit": ("centibar", "cm"),
        }
        for p in unit_parameters:
            setattr(self, p, parameters.get(p, unit_parameters[p][0]))
            if self.__dict__[p] not in unit_parameters[p]:
                raise ConfigurationError(
                    "{} must be one of {}".format(p, ", ".join(unit_parameters[p]))
                )

    def _extract_timestamp(self):
        pass

    def _get_storage_tail(self, after_timestamp):
        """Read the part of the data storage after_timestamp.

        See the docstring of the inherited method for more information.
        """
        result = []
        saveddir = os.getcwd()
        try:
            os.chdir(self.path)
            first_file = "{0.year:04}-{0.month:02}.wlk".format(after_timestamp)
            filename_regexp = re.compile(r"\d{4}-\d{2}.wlk$")
            data_files = [
                x for x in glob("*.wlk") if filename_regexp.match(x) and x >= first_file
            ]
            data_files.sort()
            for current_file in data_files:
                result.extend(self._get_tail_part(after_timestamp, current_file))
        finally:
            os.chdir(saveddir)
        return result

    def _get_tail_part(self, after_timestamp, filename):
        """Read a single wdat5 file.

        Reads the single wdat5 file "filename" for records with
        date > after_timestamp, and returns a list of records in space-delimited
        format; iso datetime first, values after.
        """
        year, month = [
            int(x) for x in os.path.split(filename)[1].split(".")[0].split("-")
        ]
        result = []
        with open(filename, "rb") as f:
            header = f.read(212)
            if header[:6] != b"WDAT5.":
                raise MeteologgerStorageReadError(
                    "File {0} does not appear to be a WDAT 5.x file".format(filename)
                )
            for day in range(1, 32):
                i = 20 + (day * 6)
                j = i + 6
                day_index = header[i:j]
                records_in_day = struct.unpack("<h", day_index[:2])[0]
                start_pos = struct.unpack("<l", day_index[2:])[0]
                for r in range(records_in_day):
                    f.seek(212 + ((start_pos + r) * 88))
                    record = f.read(88)
                    if record[0] != b"\x01"[0]:
                        continue
                    decoded_record = self.__decode_wdat_record(record)
                    timestamp = dt.datetime(
                        year=year, month=month, day=day
                    ) + dt.timedelta(minutes=decoded_record["packedtime"])
                    timestamp = self._fix_dst(timestamp)
                    if timestamp <= after_timestamp:
                        continue
                    decoded_record["timestamp"] = timestamp
                    result.append(decoded_record)
        return result

    def __decode_wdat_record(self, record):
        """Decode bytes into a dictionary."""
        result = {}

        # Read raw values
        offset = 0
        for item in self.wdat_record_format:
            fmt, name = item.split()
            result[name.lower()] = struct.unpack_from(fmt, record, offset)[0]
            offset += struct.calcsize(fmt)

        # Convert temperature
        for x in ["outsidetemp", "hioutsidetemp", "lowoutsidetemp", "insidetemp"]:
            result[x] = (
                result[x] / 10.0
                if self.temperature_unit == "F"
                else ((result[x] / 10.0) - 32) * 5 / 9.0
            )

        # Convert pressure
        result["barometer"] = (
            result["barometer"] / 1000.0
            if self.pressure_unit == "inch Hg"
            else result["barometer"] / 1000.0 * 25.4 * 1.33322387415
        )

        # Convert humidity
        for x in ["outsidehum", "insidehum"]:
            result[x] = result[x] / 10.0

        # Convert rain
        rain_collector_type = result["rain"] & 0xF000
        rain_clicks = result["rain"] & 0x0FFF
        depth_per_click = {
            0x0000: 0.1 * 25.4,
            0x1000: 0.01 * 25.4,
            0x2000: 0.2,
            0x3000: 1.0,
            0x6000: 0.1,
        }[rain_collector_type]
        depth = depth_per_click * rain_clicks
        result["rain"] = depth / 25.4 if self.rain_unit == "inch" else depth
        rate = result["hirainrate"] * depth_per_click
        result["hirainrate"] = rate / 25.4 if self.rain_unit == "inch" else rate

        # Convert wind speed
        def convert_wind_speed(x):
            return (
                x / 10.0
                if self.wind_speed_unit == "mph"
                else x / 10.0 * 1609.344 / 3600
            )

        result["windspeed"] = convert_wind_speed(result["windspeed"])
        result["hiwindspeed"] = convert_wind_speed(result["hiwindspeed"])

        # Convert wind direction
        for x in ["winddirection", "hiwinddirection"]:
            result[x] = result[x] / 16.0 * 360 if result[x] >= 0 else "NaN"

        # Convert UV index
        result["uv"] = result["uv"] / 10.0
        result["hiuv"] = result["hiuv"] / 10.0

        # Convert evapotranspiration
        result["et"] = result["et"] / 1000.0
        if self.rain_unit == "inch":
            result["et"] *= 25.4

        # Convert matric potential
        for i in range(1, 7):
            varname = "soilmoisture" + str(i)
            result[varname] = (
                result[varname]
                if self.matric_potential_unit == "centibar"
                else result[varname] / 9.80638
            )

        # Convert extraTemp etc.
        extratemps = [
            "extratemp1",
            "extratemp2",
            "extratemp3",
            "extratemp4",
            "extratemp5",
            "extratemp6",
            "extratemp7",
            "soiltemp1",
            "soiltemp2",
            "soiltemp3",
            "soiltemp4",
            "soiltemp5",
            "soiltemp6",
            "leaftemp1",
            "leaftemp2",
            "leaftemp3",
            "leaftemp4",
        ]
        for x in extratemps:
            result[x] = (
                result[x] - 90
                if self.temperature_unit == "F"
                else ((result[x] - 90) - 32) * 5 / 9.0
            )

        return result

    def _extract_value_and_flags(self, ts_id, record):
        for v, tid in self.variables.items():
            if tid == ts_id:
                break
        return (record[v], "")


class MeteologgerStorage_odbc(MeteologgerStorage_simple):
    def get_required_parameters(self):
        return super().get_required_parameters() | {"table", "date_sql", "data_columns"}

    def get_optional_parameters(self):
        return super().get_optional_parameters() | {"date_format", "decimal_separator"}

    def __init__(self, parameters, logger=None):
        super().__init__(parameters, logger)
        self.table = parameters.get("table", "")
        self.date_sql = parameters.get("date_sql", "")
        self.data_columns = parameters.get("data_columns", "").split(",")
        self.delimiter = ";"

    def _get_storage_tail(self, after_timestamp):
        """Read the part of the data storage after_timestamp.

        See the docstring of the inherited method for more information.
        """
        try:
            # In the beginning of the file we have "pyodbc = None", so normally here
            # pyodbc will be imported. But if in the meantime some unit test has patched
            # pyodbc, no importing will be done.
            global pyodbc
            if pyodbc is None:
                import pyodbc
        except ImportError:
            self.logger.error("Install pyodbc to use odbc format")
            raise
        sql = """SELECT {} + ';' + {} FROM "{}" ORDER BY -id""".format(
            self.date_sql,
            " + ';' + ".join(['"{}"'.format(x) for x in self.data_columns]),
            self.table,
        )
        result = []
        connection = pyodbc.connect(self.path)
        cursor = connection.cursor()
        cursor.execute(sql)
        for row in cursor:  # Iterable cursor is a pyodbc feature
            line = row[0]  # Our SQL returns a single string
            self.logger.debug(line)
            date = self._extract_timestamp(line).replace(second=0)
            date = self._fix_dst(date)
            self.logger.debug("Date: " + date.isoformat())
            if date <= after_timestamp:
                break
            result.append({"timestamp": date, "line": line})
        result.reverse()
        return result



#!/usr/bin/env python

import os
import re

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

requirements = [
    "Click>=7.0,<8",
    "iso8601",
    "htimeseries>=1,<2",
    "simpletail",
    "enhydris-api-client>=2,<3",
]

setup_requirements = []

test_requirements = ["pyfakefs"]


def get_version():
    scriptdir = os.path.dirname(os.path.abspath(__file__))
    init_py_path = os.path.join(scriptdir, "loggertodb", "__init__.py")
    with open(init_py_path) as f:
        return re.search(r'^__version__ = "(.*?)"$', f.read(), re.MULTILINE).group(1)


setup(
    author="Antonis Christofides",
    author_email="antonis@antonischristofides.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="Insert meteorological station data to Enhydris",
    entry_points={"console_scripts": ["loggertodb=loggertodb.cli:main"]},
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="loggertodb",
    name="loggertodb",
    packages=find_packages(include=["loggertodb"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/aptiko/loggertodb",
    version=get_version(),
    zip_safe=False,
)



import datetime as dt
import os
import textwrap

from pyfakefs.fake_filesystem_unittest import TestCase

from loggertodb.exceptions import MeteologgerStorageReadError
from loggertodb.meteologgerstorage import (
    MeteologgerStorage_simple,
    MultiTextFileMeteologgerStorage,
)


class DummyMultiTextFileMeteologgerStorage(MultiTextFileMeteologgerStorage):
    def _extract_timestamp(self, line):
        return dt.datetime.strptime(line[:16], "%Y-%m-%d %H:%M")

    def _get_item_from_line(self, line, seq):
        line_items = line.strip().split(",")[1:]
        item = line_items[seq - 1]
        item_items = item.split()
        return float(item_items[0]), " ".join(item_items[1:])


class GetStorageTailTestCase(TestCase):
    use_headers_in_files = False

    def setUp(self):
        self.setUpPyfakefs()
        self.meteologger_storage = self._get_meteologger_storage()
        self._create_files()

    def _get_meteologger_storage(self):
        parms = {
            "station_id": 1334,
            "path": "/foo/bar?",
            "storage_format": "dummy",
            "fields": "5, 6",
            "null": "NULL",
        }
        if self.use_headers_in_files:
            parms["ignore_lines"] = "Date"
        return DummyMultiTextFileMeteologgerStorage(parms)

    def _create_files(self):
        self._create_test_file("/foo/bar1", 2018)
        self._create_test_file("/foo/bar2", 2019)
        self._create_test_file("/foo/bar3", 2017)

    def _create_test_file(self, pathname, year):
        headers = "Date,value1,value2\n" if self.use_headers_in_files else ""
        self.fs.create_file(
            pathname,
            contents=textwrap.dedent(
                """\
                {}
                {}-02-28 17:20,42.1,24.2
                {}-02-28 17:30,42.2,24.3
                """.format(
                    headers, year, year
                )
            ),
        )

    def test_get_storage_tail_from_last_file(self):
        self.result = self.meteologger_storage._get_storage_tail(
            dt.datetime(2019, 2, 28, 17, 20)
        )
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                }
            ],
        )

    def test_get_storage_tail_from_last_but_one_file(self):
        self.result = self.meteologger_storage._get_storage_tail(
            dt.datetime(2018, 2, 28, 17, 20)
        )
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2018, 2, 28, 17, 30),
                    "line": "2018-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 20),
                    "line": "2019-02-28 17:20,42.1,24.2\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
            ],
        )

    def test_get_storage_tail_from_all_files(self):
        self.result = self.meteologger_storage._get_storage_tail(
            dt.datetime(2016, 2, 28, 17, 20)
        )
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2017, 2, 28, 17, 20),
                    "line": "2017-02-28 17:20,42.1,24.2\n",
                },
                {
                    "timestamp": dt.datetime(2017, 2, 28, 17, 30),
                    "line": "2017-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2018, 2, 28, 17, 20),
                    "line": "2018-02-28 17:20,42.1,24.2\n",
                },
                {
                    "timestamp": dt.datetime(2018, 2, 28, 17, 30),
                    "line": "2018-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 20),
                    "line": "2019-02-28 17:20,42.1,24.2\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
            ],
        )


class GetStorageTailNoFilesTestCase(TestCase):
    def setUp(self):
        self.meteologger_storage = self._get_meteologger_storage()

    def _get_meteologger_storage(self):
        return DummyMultiTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar?",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
            }
        )

    def test_get_storage_tail_returns_empty_list(self):
        result = self.meteologger_storage._get_storage_tail(
            dt.datetime(2016, 2, 28, 17, 20)
        )
        self.assertEqual(len(result), 0)


class GetStorageTailWithHeadersTestCase(GetStorageTailTestCase):
    use_headers_in_files = True


class GetStorageTailEmptyFileTestCase(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.meteologger_storage = self._get_meteologger_storage()
        self._create_files()

    def _get_meteologger_storage(self):
        parms = {
            "station_id": 1334,
            "path": "/foo/bar?",
            "storage_format": "dummy",
            "fields": "5, 6",
            "null": "NULL",
            "ignore_lines": "Date",
        }
        return DummyMultiTextFileMeteologgerStorage(parms)

    def _create_files(self):
        self._create_test_file("/foo/bar1", 2018)
        self._create_test_file("/foo/bar2", 2019)
        self._create_test_file("/foo/bar3", None)

    def _create_test_file(self, pathname, year):
        if year is None:
            self._create_empty_test_file(pathname)
        else:
            self._create_test_file_with_records(pathname, year)

    def _create_empty_test_file(self, pathname):
        self.fs.create_file(pathname, contents="Date,value1,value2\n")

    def _create_test_file_with_records(self, pathname, year):
        self.fs.create_file(
            pathname,
            contents=textwrap.dedent(
                """\
                Date,value1,value2
                {}-02-28 17:20,42.1,24.2
                {}-02-28 17:30,42.2,24.3
                """.format(
                    year, year
                )
            ),
        )

    def test_get_entire_storage_tail(self):
        self.result = self.meteologger_storage._get_storage_tail(
            dt.datetime(1700, 1, 1, 0, 0)
        )
        self.assertEqual(len(self.result), 4)


class BadFileOrder(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.meteologger_storage = self._get_meteologger_storage()
        self._create_file()

    def _get_meteologger_storage(self):
        parms = {
            "station_id": 1334,
            "path": "/foo/bar?",
            "storage_format": "dummy",
            "fields": "5, 6",
            "null": "NULL",
            "ignore_lines": "Date",
        }
        return DummyMultiTextFileMeteologgerStorage(parms)

    def _create_file(self):
        self.fs.create_file(
            "/foo/bar1",
            contents=textwrap.dedent(
                """\
                Date,value1,value2
                2019-02-28 17:20,42.1,24.2
                2018-02-28 17:30,42.2,24.3
                """
            ),
        )

    def test_raises_value_error(self):
        msg = fr"The order of timestamps in file \{os.sep}foo\{os.sep}bar1 is mixed up."
        with self.assertRaisesRegex(ValueError, msg):
            self.meteologger_storage.get_recent_data(5, dt.datetime(1700, 1, 1, 0, 0))


class FilesWithOverlap(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.meteologger_storage = self._get_meteologger_storage()
        self._create_file1()
        self._create_file2()

    def _get_meteologger_storage(self):
        parms = {
            "station_id": 1334,
            "path": "/foo/bar?",
            "storage_format": "dummy",
            "fields": "5, 6",
            "null": "NULL",
            "ignore_lines": "Date",
        }
        return DummyMultiTextFileMeteologgerStorage(parms)

    def _create_file1(self):
        self.fs.create_file(
            "/foo/bar1",
            contents=textwrap.dedent(
                """\
                Date,value1,value2
                2018-02-28 17:20,42.1,24.2
                2019-02-28 17:30,42.2,24.3
                """
            ),
        )

    def _create_file2(self):
        self.fs.create_file(
            "/foo/bar2",
            contents=textwrap.dedent(
                """\
                Date,value1,value2
                2019-02-28 17:20,42.1,24.2
                2020-02-28 17:30,42.2,24.3
                """
            ),
        )

    def test_raises_value_error(self):
        msg = (
            fr"The timestamps in files \{os.sep}foo\{os.sep}bar1 and "
            fr"\{os.sep}foo\{os.sep}bar2 overlap."
        )
        with self.assertRaisesRegex(ValueError, msg):
            self.meteologger_storage.get_recent_data(5, dt.datetime(1700, 1, 1, 0, 0))


class FileWithBadLine(TestCase):
    def setUp(self):
        self.setUpPyfakefs()
        self.meteologger_storage = self._get_meteologger_storage()
        self._create_file()

    def _get_meteologger_storage(self):
        parms = {
            "station_id": 1334,
            "path": "/foo/bar?",
            "storage_format": "simple",
            "fields": "5, 6",
            "null": "NULL",
            "ignore_lines": "Date",
            "nfields_to_ignore": 1,
        }
        return MeteologgerStorage_simple(parms)

    def _create_file(self):
        self.fs.create_file(
            "/foo/bar1",
            contents=(
                b"id,Date,value1,value2\n"
                b"Invalid line\n"
                b"#501,2018-02-28 17:20,42.1,24.2\n"
                b"#502,2019-02-28 17:30,42.2,24.3\n"
            ),
        )

    def test_raises_error(self):
        msg = fr'\{os.sep}foo\{os.sep}bar1: "Invalid line": Malformed line'
        with self.assertRaisesRegex(MeteologgerStorageReadError, msg):
            self.meteologger_storage.get_recent_data(5, dt.datetime(1700, 1, 1, 0))



import datetime as dt
import textwrap
from collections import OrderedDict
from unittest import TestCase

import numpy as np
import pandas as pd
from pyfakefs.fake_filesystem_unittest import Patcher

from loggertodb import ConfigurationError
from loggertodb.meteologgerstorage import TextFileMeteologgerStorage


class DummyTextFileMeteologgerStorage(TextFileMeteologgerStorage):
    def _extract_timestamp(self, line):
        return dt.datetime.strptime(line[:16], "%Y-%m-%d %H:%M")

    def _get_item_from_line(self, line, seq):
        line_items = line.strip().split(",")[1:]
        item = line_items[seq - 1]
        item_items = item.split()
        return float(item_items[0]), " ".join(item_items[1:])


class CheckParametersTestCase(TestCase):
    def test_raises_error_on_fields_missing(self):
        expected_error_message = 'Parameter "fields" is required'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyTextFileMeteologgerStorage(
                {"station_id": 1334, "path": "irrelevant", "storage_format": "dummy"}
            )

    def test_raises_error_on_invalid_parameter(self):
        expected_error_message = 'Unknown parameter "hello"'
        with self.assertRaisesRegex(ConfigurationError, expected_error_message):
            DummyTextFileMeteologgerStorage(
                {
                    "station_id": 1334,
                    "path": "irrelevant",
                    "storage_format": "dummy",
                    "fields": "5, 6",
                    "hello": "world",
                }
            )

    def test_accepts_allowed_optional_parameters(self):
        DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
            }
        )


class TimeseriesIdsTestCase(TestCase):
    def test_timeseries_group_ids(self):
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "irrelevant",
                "storage_format": "dummy",
                "fields": "0, 5, 0, 6",
                "null": "NULL",
            }
        )
        self.assertEqual(meteologger_storage.timeseries_group_ids, set((5, 6)))


class ExtractValueAndFlagsTestCase(TestCase):
    def setUp(self):
        self.meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
            }
        )
        self.record = {
            "timestamp": dt.datetime(2019, 2, 28, 17, 30),
            "line": "2019-02-28 17:30,42.2 MYFLAG1 MYFLAG2,24.3\n",
        }

    def test_value_with_flags(self):
        result = self.meteologger_storage._extract_value_and_flags(5, self.record)
        self.assertAlmostEqual(result[0], 42.2)
        self.assertEqual(result[1], "MYFLAG1 MYFLAG2")

    def test_value_without_flags(self):
        result = self.meteologger_storage._extract_value_and_flags(6, self.record)
        self.assertAlmostEqual(result[0], 24.3)
        self.assertEqual(result[1], "")


class GetStorageTailTestCase(TestCase):
    def setUp(self):
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
            }
        )
        with Patcher() as patcher:
            patcher.fs.create_file(
                "/foo/bar",
                contents=textwrap.dedent(
                    """\
                    2019-02-28 17:20,42.1,24.2
                    2019-02-28 17:30,42.2,24.3
                    2019-02-28 17:40,42.3,24.4
                    """
                ),
            )
            self.result = meteologger_storage._get_storage_tail(
                dt.datetime(2019, 2, 28, 17, 20)
            )

    def test_result(self):
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 40),
                    "line": "2019-02-28 17:40,42.3,24.4\n",
                },
            ],
        )


class GetRecentDataWithAmbiguousHourTestCase(TestCase):
    def setUp(self):
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
                "timezone": "Europe/Athens",
            }
        )
        with Patcher() as patcher:
            patcher.fs.create_file(
                "/foo/bar",
                contents=textwrap.dedent(
                    """\
                    2018-10-28 02:50,42.1,24.2
                    2018-10-28 03:00,42.2,24.3
                    2018-10-28 03:10,42.3,24.4
                    2018-10-28 03:20,42.4,24.5
                    2018-10-28 03:30,42.5,24.6
                    2018-10-28 03:40,42.6,24.7
                    2018-10-28 03:50,42.7,24.8
                    2018-10-28 03:00,42.8,24.9
                    2018-10-28 03:10,42.9,25.0
                    2018-10-28 03:20,43.0,25.1
                    2018-10-28 03:30,43.1,25.2
                    2018-10-28 03:40,43.2,25.3
                    2018-10-28 03:50,43.3,25.4
                    2018-10-28 04:00,43.4,25.5
                    2018-10-28 04:10,43.5,25.6
                    2019-10-27 02:50,42.1,24.2
                    2019-10-27 03:00,42.2,24.3
                    2019-10-27 03:10,42.3,24.4
                    2019-10-27 03:20,42.4,24.5
                    2019-10-27 03:30,42.5,24.6
                    2019-10-27 03:40,42.6,24.7
                    2019-10-27 03:50,42.7,24.8
                    2019-10-27 03:00,42.8,24.9
                    2019-10-27 03:10,42.9,25.0
                    2019-10-27 03:20,43.0,25.1
                    2019-10-27 03:30,43.1,25.2
                    2019-10-27 03:40,43.2,25.3
                    2019-10-27 03:50,43.3,25.4
                    2019-10-27 04:00,43.4,25.5
                    2019-10-27 04:10,43.5,25.6
                    """
                ),
            )
            self.result = meteologger_storage.get_recent_data(
                5, dt.datetime(2018, 10, 27, 1, 0)
            )

    def test_result(self):
        v = np.arange(421, 436) / 10.0
        expected_values = np.concatenate((v, v))
        expected_flags = [""] * 30
        expected_dates = np.concatenate(
            (
                pd.date_range(
                    start=dt.datetime(2018, 10, 28, 1, 50),
                    end=dt.datetime(2018, 10, 28, 4, 10),
                    freq="10min",
                ),
                pd.date_range(
                    start=dt.datetime(2019, 10, 27, 1, 50),
                    end=dt.datetime(2019, 10, 27, 4, 10),
                    freq="10min",
                ),
            )
        )
        expected_result = pd.DataFrame(
            data=OrderedDict([("value", expected_values), ("flags", expected_flags)]),
            index=expected_dates,
            dtype=object,
        )
        pd.testing.assert_frame_equal(self.result, expected_result)


class IgnoreLinesTestCase(TestCase):
    def setUp(self):
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
                "ignore_lines": "ignore( this)? line",
            }
        )
        with Patcher() as patcher:
            patcher.fs.create_file(
                "/foo/bar",
                contents=textwrap.dedent(
                    """\
                    ignore line
                    2019-02-28 17:20,42.1,24.2
                    ignore this line
                    2019-02-28 17:30,42.2,24.3
                    yes, really ignore line man!
                    2019-02-28 17:40,42.3,24.4
                    """
                ),
            )
            self.result = meteologger_storage._get_storage_tail(
                dt.datetime(2019, 2, 28, 17, 20)
            )

    def test_result(self):
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 40),
                    "line": "2019-02-28 17:40,42.3,24.4\n",
                },
            ],
        )


class EncodingTestCase(TestCase):
    def setUp(self):
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
                "ignore_lines": "n' est pas",
                "encoding": "iso-8859-1",
            }
        )
        with Patcher() as patcher:
            patcher.fs.create_file(
                "/foo/bar",
                encoding="iso-8859-1",
                contents=textwrap.dedent(
                    """\
                    2019-02-28 17:20,42.1,24.2
                    Cette ligne n' est pas très importante
                    2019-02-28 17:30,42.2,24.3
                    2019-02-28 17:40,42.3,24.4
                    """
                ),
            )
            self.result = meteologger_storage._get_storage_tail(
                dt.datetime(2019, 2, 28, 17, 20)
            )

    def test_result(self):
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 40),
                    "line": "2019-02-28 17:40,42.3,24.4\n",
                },
            ],
        )


class EncodingErrorsTestCase(TestCase):
    def setUp(self):
        # We use an iso-8859-1-encoded file but without declaring the encoding in the
        # parameters, so when it attempts to read it it will try utf8. It should forgive
        # the error
        meteologger_storage = DummyTextFileMeteologgerStorage(
            {
                "station_id": 1334,
                "path": "/foo/bar",
                "storage_format": "dummy",
                "fields": "5, 6",
                "null": "NULL",
                "ignore_lines": "n' est pas",
            }
        )
        with Patcher() as patcher:
            patcher.fs.create_file(
                "/foo/bar",
                encoding="iso-8859-1",
                contents=textwrap.dedent(
                    """\
                    2019-02-28 17:20,42.1,24.2
                    Cette ligne n' est pas très importante
                    2019-02-28 17:30,42.2,24.3
                    2019-02-28 17:40,42.3,24.4
                    """
                ),
            )
            self.result = meteologger_storage._get_storage_tail(
                dt.datetime(2019, 2, 28, 17, 20)
            )

    def test_result(self):
        self.assertEqual(
            self.result,
            [
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 30),
                    "line": "2019-02-28 17:30,42.2,24.3\n",
                },
                {
                    "timestamp": dt.datetime(2019, 2, 28, 17, 40),
                    "line": "2019-02-28 17:40,42.3,24.4\n",
                },
            ],
        )
